/****************************************************************************
 *              Copyright 2002, Trustees of Boston University.              *
 *                          All Rights Reserved.                            *
 *                                                                          *
 * Author:      Shudong Jin                                                 *
 ****************************************************************************/





/****************************************************************************
 * This is the main program of live streaming media workload generator.     *
 ****************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "config.h"	/* parameters, system configurations */

#include "zipf.h"	/* zipf variate */
#include "lognormal.h"	/* lognormal variate */
#include "pareto.h"	/* pareto variate */
#include "diurnal.h"	/* diurnal arrival patterns */
#include "vbr.h"	/* self-similar vbr lib functions */


/****************************************************************************
 * SUMMARY OF THE WORKLOAD GENERATOR:                                       *
 * Client requests are grouped into sessions that arrive closely in time.   *
 * The session arrivals are generated independent, but the clients have     *
 * skewed interests (zipf-like). So the sessions are assigned to clients    *
 * non-uniformly. Each session is an ON-OFF process, where an ON period is  *
 * the transfer of a live feed (called an object), and an OFF period is     *
 * the silent period (thinkign time, pause etc) of the client. The objects  *
 * may have different popularity (zipf-like). Each object has its variable  *
 * bit rate with self-similar property and potentially high variability.    *
 ****************************************************************************/

struct session {
	int	client;		/* which client? */
	int	transfers;	/* how many transfers in this session? */
	int	**object;	/* a transfer of which object? */
	int	*ontime;	/* ON length */
	int	*offtime;	/* OFF length */
	double	clock;		/* session start time */
};

/*
 * Only for accuracy
 */
static int random_roundup(double d)
{
	return ((int)d) + (d-(int)d  > drand48());
}

/*
 * Written for qsort, sorting the sessions by their start time
 */
static int comp2Sessions(const void *s1, const void *s2)
{
	if (((struct session*)s1)->clock < ((struct session*)s2)->clock)
		return -1;
	else return 1;
}

/*************************************************************************
 * This routine generates client sessions and transfers                  *
 *************************************************************************/
static struct session* Generating_sessions(int* N, int C, int S, struct session Session[], int NUM_ATTRIBUTES, float* alpha)
{
	int i, j, k, n, count, tclock;
        struct session* sess;

	count = 0;
        tclock = 0;

	/*
	 * For each session
	 */
	for(i = 0; i < S; i++) {
		/*
		 * Choose a client following zipf-like popularity
		 */
		Session[i].client = zipf_variate(C, CLIENT_ALPHA);
		/*
		 * #transfers in each session follows Pareto
		 */
		Session[i].transfers = random_roundup(
					pareto_variate(1, TRANSFERS_ALPHA));
		n = Session[i].transfers;
                count += n;

		Session[i].object = (int**)malloc(NUM_ATTRIBUTES*sizeof(int));
		for(j = 0; j < NUM_ATTRIBUTES; j++) {
		  Session[i].object[j] = (int*)malloc(n*sizeof(int));
		}
		Session[i].ontime = (int*)malloc(n*sizeof(int));
		Session[i].offtime = (int*)malloc(n*sizeof(int));
		/*
		 * for each transfer
		 */
		for(j = 0; j < n; j++) {
		  /*
		   * Choose an object following zipf-like popularity
		   */
		  for(k = 0; k < NUM_ATTRIBUTES; k++) {
		    Session[i].object[k][j] = zipf_variate(N[k], alpha[k]);
		  }

		  /*
		   * Generate lognormal ON length
		   */
		  Session[i].ontime[j] = random_roundup(
							lognormal_variate(
									  TRANSFER_ON_LOG_MEAN,
									  TRANSFER_ON_LOG_DERIVATION));
		  /*
		   * Generate lognormal OFF length
		   */
		  Session[i].offtime[j] = random_roundup(
							 lognormal_variate(
									   TRANSFER_OFF_LOG_MEAN,
									   TRANSFER_OFF_LOG_DERIVATION));
		}
		/*
		 * Generate session arrival time, with diurnal patterns
		 */
		Session[i].clock = diurnal_variate(CONFIG_DIURNAL, drand48());
		Session[i].clock *= WORKLOAD_DURATION;
	}

        qsort(Session, S, sizeof(struct session), comp2Sessions);

        sess = (struct session *) malloc(count * sizeof(struct session));
        count = 0;

	for(i = 0; i < S; i++) {
            tclock = Session[i].clock;
            for(j = 0; j < Session[i].transfers; j++) {

                sess[count].client = Session[i].client;

                sess[count].transfers =  1;
                sess[count].object = (int**) malloc(NUM_ATTRIBUTES*sizeof(int));

                for(k = 0; k < NUM_ATTRIBUTES; k++) {
                    sess[count].object[k] = (int*) malloc(sizeof(int));
                }

                sess[count].ontime = (int*) malloc(sizeof(int));
                sess[count].offtime = (int*) malloc(sizeof(int));

                for(k = 0; k < NUM_ATTRIBUTES; k++) {
                    sess[count].object[k][0] = Session[i].object[k][j];
                }

		sess[count].ontime[0] = Session[i].ontime[j];
                sess[count].offtime[0] = Session[i].offtime[j];

                sess[count].clock = tclock;

                tclock += Session[i].ontime[j];
                tclock += Session[i].offtime[j];

                count++;
            }
        }

	qsort(sess, count, sizeof(struct session), comp2Sessions);
        return sess;
}

static void output_sessions(struct session Session[], int S, int NUM_ATTRIBUTES)
{
	int i, j, k, time;
	for(i = 0; i < S; i++) {
		int clock = Session[i].clock;
		for(j = 0; j < Session[i].transfers; j++) {
			printf("%d\t%d\t", Session[i].client, clock);

			for(k = 0; k < NUM_ATTRIBUTES; k++) {
			    printf("%d\t", Session[i].object[k][j]);
			}

			printf("\n");

			clock += Session[i].ontime[j];
			clock += Session[i].offtime[j];
		}
	}
}

main(int argc, char *argv[])
{
	int		C, S, ATTR;
	int		i;
        float           *alpha;
        int             *N;
	int		*Size;
	struct session	* Session;
	unsigned short	**Frame;

	if (argc < 6) {
		printf("Usage: %s {#clients} {#sessions} {#attributes} {#attr1_alpha} <#attr2_alpha>... {#attr1_objects} <#attr2_objects>...\n",argv[0]);
		exit(0);
	}

	C = atoi(argv[1]);
	if (C < 1 || C > 10000000) {
		printf("Number of clients too small or too large\n");
		exit(0);
	}

	S = atoi(argv[2]);
	if (S <= C || S > 1000*C) {
		printf("Number of sessions too small or too large\n");
		exit(0);
	}

	ATTR = atoi(argv[3]);
	if(ATTR <= 0 || ATTR > 10) {
	  printf("Number of attributes too small or too large\n");
	  exit(0);
	}

	if(argc < 4+ATTR+ATTR) {
	  printf("Missing alpha value for object attributes\n");
	  exit(0);
	}

	alpha = (float *) malloc(ATTR * sizeof(float));
	for(i = 0; i < ATTR; i++) {
	  alpha[i] = atof(argv[4+i]);
	}

        N = (int *) malloc(ATTR * sizeof(int));
        for(i = 0; i < ATTR; i++) {
          N[i] = atoi(argv[4+ATTR+i]);
        }

	srand48(getpid());

	/*
	 * Generating a sequence of client sessions
	 */
	Session = (struct session *) malloc(S * sizeof(struct session));
	Session = Generating_sessions(N, C, S, Session, ATTR, alpha);

	output_sessions(Session, S, ATTR);
}
