/****************************************************************************
 *              Copyright 2001, Trustees of Boston University.              *
 *                          All Rights Reserved.                            *
 *                                                                          *
 * Author:      Shudong Jin                                                 *
 ****************************************************************************/




#include <stdlib.h>
#include <math.h>
#include "vbr.h"



/****************************************************************************
 * This routine generates frame sizes which exhibit self-similarity and     *
 * follow Gaussian (normal) marginal distribution. Parameter "H" is the     *
 * target Hurst parameter of the self-similarity. Parameters "u" and "d"    *
 * are the mean and derivation of the Gaussian distribution, respectively.  *
 * The routine generates "frames" frame sizes which are returned in vector  *
 * "framesize". No maximum or minimum framesizes are specified.             *
 ****************************************************************************/
static void generateFBM(int frames, double H, double Y[]);
static double generateFGN(int frames, double Y[]);
static void transformGaussian(double u1, double d1, double u2, double d2,
			int frames, double Y[], unsigned short framesize[]);
void generate_VBR_gaussian(double H, double u, double d,
			int frames, unsigned short framesize[])
{
	double d1;
	double *Y = (double *)malloc((frames+1) * sizeof(double));

	generateFBM(frames+1, H, Y);
	d1 = generateFGN(frames+1, Y);
	transformGaussian(0.0, d1, u, d, frames, Y, framesize);

	free((void*)Y);
}




/****************************************************************************
 * This routine generates frame sizes which exhibit self-similarity and     *
 * follow a hybrid Gaussian/Pareto marginal distribution. Parameter "H" is  *
 * the target Hurst parameter for self-similarity. Parameters "u" and "d"   *
 * are the mean and derivation of the Gaussian distribution, respectively.  *
 * Parameter "C" specifies the cut point of the two distributions. The      *
 * parameters of the Pareto distribution is derived by equalizing both the  *
 * value and slope of the two distributions at the cut point.  The routine  *
 * generates "frames" frame sizes which are returned in vector "framesize". *
 * No maximum or minimum framesizes are specified.                          *
 ****************************************************************************/
static void generateFBM(int frames, double H, double Y[]);
static double generateFGN(int frames, double Y[]);
static void transformGaussianPareto(double u1, double d1,
			double u2, double d2, double C,
			int frames, double Y[], unsigned short framesize[]);
void generate_VBR_gaussianPareto(double H, double u, double d, double C,
			int frames, unsigned short framesize[])
{
	double d1;
	double *Y = (double *)malloc((frames+1) * sizeof(double));

	generateFBM(frames+1, H, Y);
	d1 = generateFGN(frames+1, Y);
	transformGaussianPareto(0.0, d1, u, d, C, frames, Y, framesize);

	free((void*)Y);
}


/****************************************************************************
 * This routine generates frame sizes which exhibit self-similarity and     *
 * follow a hybrid Gaussian/Pareto marginal distribution. Parameter "H" is  *
 * the target Hurst parameter for self-similarity. Parameters "u" and "d"   *
 * are the mean and derivation of the Gaussian distribution, respectively.  *
 * Parameter "a" specifies the shape of the Pareto distribution, while the  *
 * scale parameter of Pareto distribution is calculated by fitting the two  *
 * distribution. In order to do this, we need "C" to specify the cut-off    *
 * point of the two distribution, and "P" to specify the portion of mass in *
 * the Pareto tail. The routine generates "frames" frame sizes which are    *
 * returned in vector "framesize". No maximum or minimum framesizes are     *
 * specified.                                                               *
 ****************************************************************************/
static void generateFBM(int frames, double H, double Y[]);
static double generateFGN(int frames, double Y[]);
static void transformLognormalPareto(double u1, double d1,
			double u2, double d2, double C,
			int frames, double Y[], unsigned short framesize[]);
void generate_VBR_lognormalPareto(double H, double u, double d, double C,
			int frames, unsigned short framesize[])
{
	double d1;
	double *Y = (double *)malloc((frames+1) * sizeof(double));

	generateFBM(frames+1, H, Y);
	d1 = generateFGN(frames+1, Y);

	transformLognormalPareto(0.0, d1, u, d, C, frames, Y, framesize);

	free((void*)Y);
}

/****************************************************************************
 * This routine provides an estimation of the Hurst parameter, given the    *
 * "frames" frame sizes in vector "framesize". The estimation bases on time *
 * -variance plot and uses least-square fit. In order that this routine will*
 * return the result properly, the size of the frame size vector should be  *
 * large enough (>1000 for example).                                        *
 ****************************************************************************/
static double lsfit(const double X[], const double Y[], int n);
double HurstEstimation(int frames, const unsigned short framesize[])
{
	int i, j, k, t, oldt=0;
	double T;
	double sum, mean, variance, normal = 10000.0;
	double logTime[1000], logVariance[1000];
	int index = 0;

	for(T = 1; T < frames/30; T = T*1.1) {
		t = T;
		if (oldt == t) continue;
		sum = 0;
		for(i = 0; i < frames/t*t; i++) sum += framesize[i];
		mean = sum / (frames/t);
		variance = 0.0;
		for(j = 0; j < frames/t*t; j+=t) {
			int total = 0;
			double var;
			for(k = 0; k < t; k++) total += framesize[j+k];
			var = total-mean;
			variance += var * var;
		}

		if (t == 1) normal = variance;
		logTime[index] = log(t);
		logVariance[index] = log(variance/t);
		index++;
	}
	return 1 + lsfit(logTime, logVariance, index)/2;
}






/*
 * The followings are utility routines used only locally.
 */
 


static double nextGaussian(void)
{
	double r = drand48();
	r = sqrt(-2 * log(r));
	if (drand48() < 0.5) r = -r;
	return r;
}

static int Divide(int level, double leftY, double rightY,
		int frames, double H, double *Y, int p)
{
	if (p >= frames) return p;
	if (level > 1) {
		double midY = (leftY + rightY) * 0.5
			+ nextGaussian()*exp(level*H*0.693147180559945);
		level--;
		p = Divide(level, leftY, midY, frames, H, Y, p);
		p = Divide(level, midY, rightY, frames, H, Y, p);
		return p;
	} else {
		Y[p] = leftY;
		return p+1;
	}
}

/*
 * Middlepoint displacement approach.
 */
static void generateFBM(int frames, double H, double Y[])
{
	int level = log(frames) / 0.693147180559945 + 9;
	Divide(level, nextGaussian(), nextGaussian(), frames, H, Y, 0);
}

static double generateFGN(int frames, double Y[])
{
	int i;
	double variance = 0.0;
	for(i = 0; i < frames-1; i++) {
		Y[i] = Y[i+1] - Y[i];
		variance += Y[i]*Y[i];
	}
	return sqrt(variance/frames);

}

static double Gaussian(double x)
{
        return exp(-x*x/2) / sqrt(2*M_PI);
}

static double GaussianCDF(double x)
{
	double r = 0.2316419;
	double t1 = 1/(1 + r * (x>0.0?x:-x));
	double b1 = .319381530;
	double b2 =-.356563782;
	double b3 = 1.781477937;
	double b4 =-1.821255978;
	double b5 = 1.330274429;
	double t2 = t1 * t1;
	double t3 = t2 * t1;
	double t4 = t2 * t2;
	double t5 = t2 * t3;
        double R = Gaussian(x) * (b1*t1 + b2*t2+ b3*t3 + b4*t4+ b5*t5);
        if (x>=0) return 1-R; else return R;
}

static double InverseGaussianCDF(double y)
{
	double c0 = 2.515517;
	double c1 = 0.802853;
	double c2 = 0.010328;
	double d1 = 1.432788;
	double d2 = 0.189269;
	double d3 = 0.001308;
	double t, x;

	y = 1 - y;
	if (y <= 0.5) t = sqrt(-2*log(y));
	else t = sqrt(-2*log(1-y));
	x = t - (c0 + c1*t + c2*t*t) / (1 + d1*t + d2*t*t + d3*t*t*t);
	return (y<=0.5 ? x : -x);
}

static void transformGaussian(double u1, double d1,
			double u2, double d2,
			int frames, double Y[], unsigned short framesize[])
{
	int i;
	double b = 1 / (GaussianCDF((65535-u2)/d2) - GaussianCDF(-u2/d2));
	double tmp = GaussianCDF(-u2/d2);
	for(i = 0; i < frames; i++) {
		double y = GaussianCDF((Y[i]-u1)/d1);
		y = y/b + tmp;
		framesize[i] = u2 + d2 * InverseGaussianCDF(y);
	}
}

static void transformGaussianPareto(double u1, double d1,
			double u2, double d2, double C,
			int frames, double Y[], unsigned short framesize[])
{
	double tmp1, tmp2, tmp3;
	int i;
	double P = exp(-(C-u2)*(C-u2)/d2/d2/2) / sqrt(2*M_PI);
	double slopeC = -(C-u2) * P / (d2*d2*d2);
	double valueC = P / d2;
	double a = -1 - slopeC / valueC * C;
	double k = exp( log(valueC/a*exp((a+1)*log(C))) / a);

	double b = GaussianCDF((C-u2)/d2) - GaussianCDF(-u2/d2);
	b += exp(a*log(k/C));
	b = 1/b;
	/* now b is the normalizing factor */


	P = 1 - exp(a*log(k/C)) * b;
	/* now P is the portion of the body */

 	tmp1 = GaussianCDF(-u2/d2);
	tmp2 = exp(-a*log(k));
	tmp3 = exp(-a*log(C));
	for(i = 0; i < frames; i++) {
		double y = GaussianCDF( (Y[i]-u1)/d1 );
		if (y < P) {
			y = y/b + tmp1;
			framesize[i] = u2 + d2*InverseGaussianCDF(y);
		} else {
			y = (y - P) / b;
			y *= tmp2;
			y = tmp3 - y;
			y = exp(-log(y)/a);
			framesize[i] = y;
		}
	}
}

static void transformLognormalPareto(double u1, double d1,
			double u2, double d2, double C,
			int frames, double Y[], unsigned short framesize[])
{
	double tmp1, tmp2;
	int i;
	double P = exp(-(log(C)-u2)*(log(C)-u2)/d2/d2/2) / sqrt(2*M_PI)/C/d2;
	double valueC = P;
	double slopeC = P / C * (-1 - (log(C)-u2)/d2/d2);
	double a = -1 - slopeC / valueC * C;
	double k = exp( log(valueC/a*exp((a+1)*log(C))) / a);

	double b = GaussianCDF((log(C)-u2)/d2);
	b += exp(a*log(k/C));
	b = 1/b;
	/* now b is the normalizing factor */

	P = 1 - exp(a*log(k/C)) * b;
	/* now P is the portion of the body */

	tmp1 = exp(-a*log(k));
	tmp2 = exp(-a*log(C));
	for(i = 0; i < frames; i++) {
		double y = GaussianCDF( (Y[i]-u1)/d1 );
		if (y < P) {
			y = y/b;
			framesize[i] = exp(u2 + d2*InverseGaussianCDF(y));
		} else {
			y = (y - P) / b;
			y *= tmp1;
			y = tmp2 - y;
			y = exp(-log(y)/a);
			framesize[i] = y;
		}
	}
}

static double lsfit(const double X[], const double Y[], int n)
{
	int i;
	double sumx = 0.0;
	double sumy = 0.0;
	double sumxx = 0.0;
	double sumxy = 0.0;
	double D = 0.0;

	for (i = 0; i < n; i++) {
		sumx += X[i];
		sumy += Y[i];
		sumxx += (X[i] * X[i]);
		sumxy += (X[i] * Y[i]);
	}
	D = sumxx * n - sumx * sumx;
	return (sumxy * n - sumx * sumy) / D;
}
